(function()
{
    
    const {Cc, Ci}  = require("chrome");
    const log = require("ko/logging").getLogger("codeintel");
    const process = require("codeintel/process");
    const stylesheet = require("ko/stylesheet");
    const legacy = require("ko/windows").getMain().ko;
    const timers = require("sdk/timers");
    const prefs = require("ko/prefs");
    const observerSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    const isWindows = require("sdk/system").platform.toLowerCase() == "winnt";
    const pathSep = isWindows ? ";" : ":";
    const langReg = Cc['@activestate.com/koLanguageRegistryService;1'].getService(Ci.koILanguageRegistryService);
    const koEnviron = Cc["@activestate.com/koUserEnviron;1"].getService();
    const service = require("codeintel/service");
    const l = require("ko/locale").use();

    this.FEATURE_COMPLETIONS = "completions";
    this.FEATURE_CALLTIPS = "calltips";
    this.FEATURE_SYMBOLBROWSER = "symbolbrowser";
    this.FEATURE_SYMBOLLIST = "symbollist";
    this.FEATURE_SYMBOLSCOPE = "symbolscope";
    this.FEATURE_SCANNER = "scanner";
    this.FEATURE_GOTODEF = "gotodef";
    this.FEATURE_JUMPSECTIONS = "jumpsections";

    var features = [
        this.FEATURE_COMPLETIONS,
        this.FEATURE_CALLTIPS,
        this.FEATURE_SYMBOLBROWSER,
        this.FEATURE_SYMBOLLIST,
        this.FEATURE_SYMBOLSCOPE,
        this.FEATURE_SCANNER,
        this.FEATURE_GOTODEF,
        this.FEATURE_JUMPSECTIONS,
    ];

    //log.setLevel(10);

    var storage = require("ko/session-storage").get("codeintel").storage;

    var socket = storage.socket || null;
    var client = storage.client || null;

    var keepaliveTimer;
    var featuresLoaded = false;

    this.getClient = () =>
    {
        return client;
    };

    var init = () =>
    {
        if (storage.running)
        {
            loadFeatures();
        }
        else
        {
            storage.running = true;
            this.start();
        }
        stylesheet.loadGlobal("less://codeintel/skin/style/codeintel.less");
    };

    this.start = () =>
    {
        process.start();
        process.on("started", onServerReady);
        observerSvc.addObserver(stopObserver, "quit-application", false);

        require("ko/prefs").registerCategory(
           l.get("code_intelligence"),
           "chrome://codeintel/content/prefs.xul",
           "uiSettingsItem"
        );

        keepalive();
    };

    var stopObserver = { observe: this.stop };

    this.stop = () =>
    {
        log.info("Stop called");

        unloadFeatures();

        this.stopSocket();
        process.stop();

        observerSvc.removeObserver(stopObserver, "quit-application");
        
        timers.clearTimeout(keepaliveTimer);
    };

    this.stopSocket = () =>
    {
        if (socket)
            socket.close();

        socket = storage.socket = null;
        client = storage.client = null;
    };
    
    this.isFeatureEnabled = (feature) =>
    {
        return prefs.getBoolean(`codeintel.${feature}.enabled`);
    };
    
    this.getImportPaths = (prefset, language, languages) =>
    {
        var languageObj = langReg.getLanguage(language);
        if ( ! languageObj)
            return [];
        
        var importPaths = [];
        if (language in languages && languages[language].extrapaths)
        {
            var key = languages[language].extrapaths;
            importPaths = koEnviron.getEnvVar(key, prefset).split(pathSep);
        }

        // Append additional import paths, if any
        if (languageObj.importPref)
        {
            var importPref = prefset.getString(languageObj.importPref, "");
            if (importPref)
            {
                importPaths = importPaths.concat(importPref.split(pathSep));
            }
        }

        importPaths = importPaths.filter((p) => !!p);

        return importPaths;
    };
    
    var keepalive = () =>
    {
        // This function is not responsible for restarting the process,
        // just for sending the keepalive requests

        if (process.get("status") == process.STATUS_STARTED)
        {
            service.keepalive();
        }
        
        keepaliveTimer = timers.setTimeout(keepalive, prefs.getLong("codeintel.keepalive.interval"));
    };

    var loadFeatures = () =>
    {
        for (let feature of features)
        {
            if ( ! this.isFeatureEnabled(feature))
            {
                log.debug("Not loading disabled feature: " + feature);
                continue;
            }
            
            try
            {
                log.debug("Loading feature: " + feature);
                require("codeintel/feature/" + feature).start();
            } catch (e)
            {
                log.exception(e, "Exception while loading feature: " + feature);
            }
        }
    };
    
    var unloadFeatures = () =>
    {
        for (let feature of features)
        {
            try
            {
                log.debug("Unloading feature: " + feature);
                require("codeintel/feature/" + feature).stop();
            } catch (e)
            {
                log.exception(e, "Exception while unloading feature: " + feature);
            }
        }
    };
    
    var onServerReady = () =>
    {
        socket = storage.socket = require("ko/socket/tcp").open("127.0.0.1", process.get("port"));
        client = storage.client = require("ko/jsonrpc").create(socket);
        
        if (featuresLoaded)
            return;
        
        timers.setTimeout(() =>
        {
            loadFeatures();

            require("ko/dom")(require("ko/windows").getMain()).trigger("codeintel-ready");

            featuresLoaded = true;
        }, 0);
    };
    
    init();
    
}).apply(module.exports);
