/**
 * @module ko/dom
 *
 * options:
 * {
 *   <int> id: identifier for this call, used to merge repeat calls (the id otherwise gets auto generated)
 *   <boolean> cache: whether or not caching is enabled for this call
 *   <int> maxAge: maximum cache age (in ms)
 *   <string> method: the request method, one of: legacy, fallback, combined, <undefined> (regular request)
 * }
 */
(function()
{
    
    const {Cc, Ci}  = require("chrome");
    const log = require("ko/logging").getLogger("codeintel/service");
    const jsonrpc = require("ko/jsonrpc");
    const mediator = require("./service/mediator");

    // alias jsonrpc error constants, because dependant libraries should not
    // interact with jsonrpc directly
    this.ERROR_PARSING = jsonrpc.ERROR_PARSING;
    this.ERROR_INVALID_REQUEST = jsonrpc.ERROR_INVALID_REQUEST;
    this.ERROR_METHOD_NOT_FOUND = jsonrpc.ERROR_METHOD_NOT_FOUND;
    this.ERROR_INVALID_PARAMS = jsonrpc.ERROR_INVALID_PARAMS;
    this.ERROR_INTERNAL = jsonrpc.ERROR_INTERNAL;

    this.flush = mediator.flush;

    /**
     * Get completions for the given buffer
     *
     * input:
     * {
     *   <string> buf: the buffer (your code)
     *   <int> pos: caret position in the buffer
     *   <string> language: language name
     *   <string> path (optional): path of the current buffer
     *   <string> parentPath (optional): the project/workspace path that this buffer belongs to
     *   <string> importPaths (optional): additional import paths
     *   <object> env (optional): environment variables
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getCompletions = mediator.register(
        "getCompletions",
        ["buf", "pos", "path", "parentPath", "importPaths", "language", "limit"],
        { cache: true, maxAge: 200, method: "fallback" }
    );

    /**
     * Get symbols in the current buffer
     *
     * input:
     * {
     *   <string> buf: the buffer (your code)
     *   <int> line: caret line in the buffer
     *   <int> pos: caret position in the buffer (only used by legacy api)
     *   <string> indentString: string representing a single indentation for the current buffer (eg. a tab, 4 spaces) (only used by legacy api)
     *   <string> language: language name
     *   <string> sortType: how to sort the results, one of: organic, alpha
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getSymbolsInBuffer = mediator.register(
        "getSymbolsInBuffer",
        ["buf", "line", "pos", "indentString", "language", "sortType"],
        { cache: true, maxAge: 200 }
    );

    /**
     * Get the scope for the current caret position
     * ie. shows you what symbol the caret is on
     *
     * input:
     * {
     *   <string> buf: the buffer (your code)
     *   <int> line: caret line in the buffer
     *   <int> pos: caret position in the buffer (only used by legacy api)
     *   <string> language: language name
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getCaretScope = mediator.register(
        "getCaretScope",
        ["buf", "line", "pos", "language"],
        { cache: true, maxAge: 200 }
    );

    /**
     * Get supported languages
     *
     * @param {object} input empty
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getLanguages = mediator.register(
        "getLanguages", [],
        { cache: true, method: "combined" }
    );

    /**
     * Get keywords for the given language
     *
     * @param {string} language
     *
     * @returns {KoPromise}
     */
    this.getKeywords = mediator.register(
        "getKeywords", ["language"],
        { cache: true }
    );

    /**
     * Get symbols based on search query
     *
     * input:
     * {
     *   <string> query: the search query
     *   <string> path: prioritize results from this path
     *   <string> parentPath: scope results to symbols contained in this path
     *   <string> type: only return symbols of the given type
     *   <string> language: language name
     *   <int> limit: maximum number of results to return
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getSymbols = mediator.register(
        "getSymbols", ["query", "path", "parentPath", "type", "language", "limit"],
        { cache: true, maxAge: 10000 }
    );

    /**
     * Retrieve a scan summary for the given path. Used to see what an actual
     * scan would be doing.
     *
     * input:
     * {
     *   <array> paths: path to scan
     *   <int> maxDepth: maximum scanning depth
     *   <array> excludes: array of excludes (as defined in the project prefs)
     *   <int> limit: maximum number of results to return
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.scanSummary = mediator.register(
        "scanSummary", ["paths", "maxDepth", "excludes", "limit"]
    );

    /**
     * Scan the given path
     *
     * input:
     * {
     *   <array> paths: paths to scan
     *   <int> maxDepth: maximum scanning depth
     *   <array> excludes: array of excludes (as defined in the project prefs)
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.scan = mediator.register(
        "scan", ["paths", "maxDepth", "excludes"]
    );
    
    /**
     * Get the next scope relative to the caret position
     *
     * input:
     * {
     *   <string> buf: the buffer (your code)
     *   <int> line: caret line in the buffer
     *   <int> pos: caret position in the buffer (only used by legacy api)
     *   <string> language: language name
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getNextScope = mediator.register(
        "getNextScope", ["buf", "line", "language"],
        { cache: true, maxAge: 10000 }
    );
    
    /**
     * Get the prev scope relative to the caret position
     *
     * input:
     * {
     *   <string> buf: the buffer (your code)
     *   <int> line: caret line in the buffer
     *   <int> pos: caret position in the buffer (only used by legacy api)
     *   <string> language: language name
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getPrevScope = mediator.register(
        "getPrevScope", ["buf", "line", "language"],
        { cache: true, maxAge: 10000 }
    );
    
    /**
     * Get a codeintel definition of the current symbol
     *
     * input object:
     * {
     *   <string> buf: the buffer (your code)
     *   <int> pos: caret position in the buffer
     *   <string> path (optional): path of the current buffer
     *   <string> parentPath (optional): the project/workspace path that this buffer belongs to
     *   <string> importPaths (optional): additional import paths
     *   <string> language: language name
     * }
     *
     * @param {object} input
     * @param {object} options (see module section above)
     *
     * @returns {KoPromise}
     */
    this.getDefinition = mediator.register(
        "getDefinition", ["buf", "pos", "path", "parentPath", "importPaths", "language"],
        { cache: true, maxAge: 10000, method: "fallback" }
    );

    this.keepalive = mediator.register(
        "keepalive", []
    );

}).apply(module.exports);
